import logging

from allianceauth import hooks
from django.contrib.auth.models import User
from django.db import transaction
from django.db.models.signals import m2m_changed, post_save, pre_delete
from django.dispatch import Signal, receiver

from . import models

# signals go here


logger = logging.getLogger(__name__)


class hook_cache:
    all_hooks = None

    def get_hooks(self):
        if self.all_hooks is None:
            hook_array = set()
            # todo add same functionality to a report source kinds thingo
            _hooks = hooks.get_hooks("secure_group_filters")
            for app_hook in _hooks:
                for filter_model in app_hook():
                    if filter_model not in hook_array:
                        hook_array.add(filter_model)
            self.all_hooks = hook_array
        return self.all_hooks


filters = hook_cache()


def new_filter(sender, instance, created, **kwargs):
    try:
        if created:
            models.ReportDataSource.objects.create(filter_object=instance)
        else:
            # this is an updated model we dont at this stage care about this.
            pass
    except:
        logger.error("Bah Humbug")  # we failed! do something here


def rem_filter(sender, instance, **kwargs):
    try:
        models.ReportDataSource.objects.get(
            object_id=instance.pk, content_type__model=instance.__class__.__name__
        ).delete()
    except:
        logger.error("Bah Humbug")  # we failed! do something here


for _filter in filters.get_hooks():
    post_save.connect(new_filter, sender=_filter)
    pre_delete.connect(rem_filter, sender=_filter)


def purge_cached_reports(sender, instance, created, **kwargs):
    """
    Clear any cached data if a report gets updated/changed/what ever...
    """
    if not created:
        models.ReportResults.objects.filter(report_id=instance.id).delete()


post_save.connect(purge_cached_reports, sender=models.Report)



import datetime
import json
import logging
import os
from collections import defaultdict

from allianceauth.authentication.models import (CharacterOwnership, State,
                                                UserProfile)
from allianceauth.eveonline.evelinks import dotlan, eveimageserver, zkillboard
from allianceauth.eveonline.models import (EveAllianceInfo, EveCharacter,
                                           EveCorporationInfo)
from allianceauth.notifications import notify
from django.contrib.auth.models import User
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from solo.models import SingletonModel

logger = logging.getLogger(__name__)


class AuthReportsConfiguration(SingletonModel):

    holding_corps = models.ManyToManyField(EveCorporationInfo)
    states_to_include = models.ManyToManyField(State)

    class Meta:
        verbose_name = "Auth Reports Configuration"
        permissions = (
            ('basic_access',
             'Can access reports module'),
            ('holding_corps',
             'Can access configured holding corp reports.'),
            ('own_corp',
             'Can access own corporations reports.'),
            ('own_alliance',
             'Can access own alliances reports.'),
            ('own_state',
             'Can access own states reports.'),
            ('restricted_reports',
             'Can access restricted reports.'),
        )

        default_permissions = []

    def holding_corp_id_qs(self):
        return self.holding_corps.all().values('corporation_id')

    def member_corp_id_qs(self):
        return UserProfile.objects.filter(
            state__in=self.states_to_include.all()
        ).values_list("main_character__corporation_id", flat=True)

    def visible_corps_for_user(self, user):
        _q = EveCorporationInfo.objects.filter(
            (
                models.Q(
                    corporation_id__in=self.member_corp_id_qs()
                ) |
                models.Q(
                    corporation_id__in=self.holding_corp_id_qs()
                )
            )
        )

        # superusers get all visible
        if user.is_superuser:
            logger.debug('Returning all corps for superuser %s.' % user)
            return _q.all()

        if user.has_perm('authstats.global_corp_manager'):
            logger.debug('Returning all corps for %s.' % user)
            return _q.all()

        try:
            char = user.profile.main_character
            assert char
            # build all accepted queries
            queries = []
            if user.has_perm('authstats.own_corp'):
                print("corp")
                queries.append(
                    models.Q(corporation_id=char.corporation_id))

            if user.has_perm('authstats.own_alliance'):
                print("alli")
                if char.alliance_id is not None:
                    queries.append(
                        models.Q(alliance__alliance_id=char.alliance_id))
                else:
                    queries.append(
                        models.Q(corporation_id=char.corporation_id))

            if user.has_perm('authstats.own_state'):
                print("state")
                queries.append(
                    models.Q(
                        corporation_id__in=EveCharacter.objects.filter(
                            character_ownership__user__profile__state=user.profile.state
                        ).values("corporation_id").distinct()
                    )
                )

            if user.has_perm('authstats.holding_corps'):
                print("holding")
                queries.append(
                    models.Q(corporation_id__in=self.holding_corp_qs()))
            logger.debug('%s queries for user %s visible corporations.' %
                         (len(queries), user))

            # filter based on queries
            if len(queries) == 0:
                return _q.none()
            query = queries.pop()
            for q in queries:
                query |= q
            return _q.filter(query)

        except AssertionError:
            logger.debug(
                'User %s has no main character. Nothing visible.' % user)
            return _q.none()


class ReportDataSource(models.Model):
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, editable=False
    )
    object_id = models.PositiveIntegerField(editable=False)
    filter_object = GenericForeignKey("content_type", "object_id")

    def __str__(self):
        try:
            return f"{self.filter_object.name}: {self.filter_object.description}"
        except:
            return f"Error: {self.content_type.app_label}:{self.content_type} {self.object_id}"

    class Meta:
        default_permissions = []


class Report(models.Model):
    name = models.CharField(max_length=250)
    report_fields = models.ManyToManyField(
        ReportDataSource, through='ReportDataThrough')
    restricted = models.BooleanField(default=False)
    show_character_image = models.BooleanField(default=True)

    def __str__(self):
        return self.name


class ReportDataThrough(models.Model):
    # through fields
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    report_Data_source = models.ForeignKey(
        ReportDataSource, on_delete=models.CASCADE)

    # report fields
    rank = models.IntegerField(
        default=5, help_text="Order the field will be show in. Lowest First.")
    header = models.CharField(
        max_length=250, help_text="Column header show to the user.")
    long_descriptions = models.CharField(
        max_length=500, help_text="Long description shown to the user on hover of column name.", null=True, default=None, blank=True)

    pass_fail_aggregate = models.BooleanField(
        default=False, help_text="Show an aggregate of how many members pass this field in the header.")
    checkbox_only = models.BooleanField(
        default=False, help_text="Show this columns data as pass/fail checkboxes only.")
    allow_sort = models.BooleanField(
        default=True, help_text="Allow sorting of this column in the UI.")

    class Meta:
        verbose_name = "Report Field"
        verbose_name_plural = verbose_name + "s"

    def __str__(self):
        return f"{self.header}"


class ReportResults(models.Model):
    corporation = models.ForeignKey(
        EveCorporationInfo, on_delete=models.CASCADE)
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    results = models.TextField()
    last_update = models.DateTimeField(auto_now=True)


# Data Fields

class FieldBase(models.Model):

    name = models.CharField(max_length=500)
    description = models.CharField(max_length=500)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.name}: {self.description}"


class ReportFieldAlts(FieldBase):
    """
        List all alts for display in report
         Options:
          - In corp only?
    """
    show_out_of_corp_alts = models.BooleanField(default=True)

    def process_field(self, users):
        co = CharacterOwnership.objects.filter(user__in=users).order_by('character__character_name').values_list(
            'user__id',
            'character__character_name',
            'character__character_id',
            "character__corporation_name",
            "character__corporation_id",
            "character__alliance_name",
            "character__alliance_id",
            "character__alliance_ticker",
            "character__corporation_ticker",
        )

        chars = defaultdict(list)
        for c in co:
            _char_url = zkillboard.character_url(c[2])
            _corp_url = zkillboard.corporation_url(c[4])
            _alli_url = zkillboard.alliance_url(c[6] if c[6] else 0)
            _str = f"<tr><td><a href='{_char_url}'>{c[1]}</a></td><td><a href='{_corp_url}'>{c[3]}</a></td>"
            if c[5]:
                _str += f"<td><a href='{_alli_url}'>{c[5]}</a></td></tr>"
            else:
                _str += f"<td></td></tr>"
            chars[c[0]].append(_str)

        output = defaultdict(lambda: {"message": "", "check": False})
        for c, char_list in chars.items():
            _msg = "".join(char_list)
            _msg = (f"<table class='table table-hover' style='white-space: nowrap;'>"
                    f"<tr><th>Character</th><th>Corporation</th><th>Alliance</th></tr>{_msg}</table>")
            output[c] = {"message": _msg, "check": True}
        return output

    class Meta:
        verbose_name = "Report Field: Character Alts"
        verbose_name_plural = verbose_name



# Generated by Django 4.0.10 on 2023-06-20 10:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authentication', '0020_userprofile_language_userprofile_night_mode'),
        ('authstats', '0002_reportdatathrough_checkbox_only'),
    ]

    operations = [
        migrations.AddField(
            model_name='authreportsconfiguration',
            name='states_to_include',
            field=models.ManyToManyField(to='authentication.state'),
        ),
    ]



from __future__ import annotations

import json
import logging
from datetime import timedelta
from typing import List

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from django.conf import settings
from django.contrib.auth.models import User
from django.db.models import QuerySet
from django.utils import timezone
from esi.errors import TokenExpiredError
from esi.models import Token
from ninja import Field, NinjaAPI, Schema
from ninja.pagination import LimitOffsetPagination, paginate
from ninja.security import django_auth
from ninja.types import DictStrAny

from authstats import schema
from authstats.tasks import run_report_for_corp

from . import providers
from .models import AuthReportsConfiguration, Report, ReportResults

logger = logging.getLogger(__name__)


class Paginator(LimitOffsetPagination):
    class Input(Schema):
        limit: int = Field(1000000, ge=1)
        offset: int = Field(0, ge=0)

    def paginate_queryset(
        self,
        queryset: QuerySet,
        pagination: Input,
        **params: DictStrAny,
    ) -> any:
        offset = pagination.offset
        limit: int = pagination.limit
        return {
            "items": queryset[offset: offset + limit],
            "count": self._items_count(queryset),
        }  # noqa: E203


api = NinjaAPI(title="Auth Stats API", version="0.0.3",
               urls_namespace='authstats:api', auth=django_auth, csrf=True)  # ,
# openapi_url=settings.DEBUG and "/openapi.json" or "")


@api.get(
    "/get_report/{report_id}/{corp_id}",
    response={200: dict, 403: str},
    tags=["Report"]
)
def get_report_for_corp(request, report_id: int, corp_id: int):
    if not AuthReportsConfiguration.get_solo().visible_corps_for_user(request.user).filter(corporation_id=corp_id).exists():
        return 403, f"Access Denied to get_report_for_corp {report_id}/{corp_id} for {request.user}"
    try:
        report = ReportResults.objects.get(
            corporation__corporation_id=corp_id, report_id=report_id)
        if report.report.restricted and not request.user.has_perms("authstats.restricted_reports"):
            return 403, f"Access Denied to get_report_for_corp {report_id}/{corp_id} for {request.user}"
        if report.last_update < timezone.now() - timedelta(hours=1):
            run_report_for_corp.delay(corp_id, report_id)
        return json.loads(report.results)
    except ReportResults.DoesNotExist:
        corp = EveCorporationInfo.objects.get(corporation_id=corp_id)
        report = Report.objects.get(id=report_id)
        # send task!
        run_report_for_corp.delay(corp_id, report_id)
        return {"report": {"name": f"{report.name}",
                           "corporation": corp.corporation_name},
                "members": 0, "unknowns": 0}


@api.get(
    "/get_corps",
    response={200: List[schema.Corporation]},
    tags=["Report"]
)
def get_corps(request):
    return AuthReportsConfiguration.get_solo().visible_corps_for_user(request.user)


@api.get(
    "/get_reports",
    response={200: List[schema.Report], 403: str},
    tags=["Report"]
)
def get_reports(request):
    if not request.user.has_perm("authstats.basic_access"):
        return 403, f"Access Denied to get_reports for {request.user}"

    _reports = Report.objects.all()
    if not request.user.has_perms("authstats.restricted_reports"):
        _reports = _reports.filter(restricted=False)
    return _reports


@api.get(
    "/get_unknowns/{corp_id}",
    response={200: dict, 403: str},
    tags=["Report"]
)
def get_orphans_for_corp(request, corp_id: int):
    if not AuthReportsConfiguration.get_solo().visible_corps_for_user(request.user).filter(corporation_id=corp_id).exists():
        return 403, f"Access Denied to get_orphans_for_corp for {request.user}"

    corp = EveCorporationInfo.objects.get(corporation_id=corp_id)

    mains = User.objects.filter(
        profile__main_character__corporation_id=corp_id)

    known_character_ids = EveCharacter.objects.filter(
        corporation_id=corp_id, character_ownership__isnull=False).values_list("character_id", flat=True)
    names = []
    try:
        scopes = "esi-corporations.read_corporation_membership.v1"
        token = Token.objects.filter(
            character_id__in=EveCharacter.objects.filter(
                corporation_id=corp_id).values('character_id')).require_scopes(scopes)
        if not token.exists():
            raise Exception("No Tokens")
        unknown_member_list = set(providers.esi.client.Corporation.get_corporations_corporation_id_members(
            corporation_id=corp_id, token=token.first().valid_access_token()).results())
        for m in known_character_ids:
            try:
                unknown_member_list.remove(m)
            except KeyError:
                pass
        names = providers.esi.client.Universe.post_universe_names(
            ids=list(unknown_member_list)).results()
    except Exception as e:
        logger.exception(e)

    return {"report": {"name": "Orphan Characters",
                       "corporation": corp.corporation_name},
            "members": mains.count(), "data": names}



import datetime
import json
import logging

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.eveonline.providers import provider as eve_names
from allianceauth.services.tasks import QueueOnce
from celery import chain, shared_task
from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import timezone
from esi.errors import TokenExpiredError
from esi.models import Token
from requests.adapters import MaxRetryError

from . import app_settings, providers
from .models import Report, ReportDataThrough, ReportResults

TZ_STRING = "%Y-%m-%dT%H:%M:%SZ"

logger = logging.getLogger(__name__)


@shared_task(bind=True, base=QueueOnce, max_retries=None)
def run_report_for_corp(self, corp_id, report_id):
    report = Report.objects.get(id=report_id)
    corp = EveCorporationInfo.objects.get(corporation_id=corp_id)

    fields = ReportDataThrough.objects.filter(report=report).order_by("rank")

    mains = User.objects.filter(
        profile__main_character__corporation_id=corp_id)

    known_character_ids = EveCharacter.objects.filter(
        corporation_id=corp_id, character_ownership__isnull=False).values_list("character_id", flat=True)
    unknown_member_list = []

    try:
        scopes = "esi-corporations.read_corporation_membership.v1"
        token = Token.objects.filter(
            character_id__in=EveCharacter.objects.filter(
                corporation_id=corp_id).values('character_id')).require_scopes(scopes)
        if not token.exists():
            raise Exception("No Tokens")
        unknown_member_list = set(providers.esi.client.Corporation.get_corporations_corporation_id_members(
            corporation_id=corp_id, token=token.first().valid_access_token()).results())
        for m in known_character_ids:
            try:
                unknown_member_list.remove(m)
            except KeyError:
                pass
    except Exception as e:
        logger.exception(e)

    output = {"report": {"name": report.name,
                         "corporation": corp.corporation_name},
              "headers": {},
              "data": {},
              "members": mains.count(),
              "unknowns": len(unknown_member_list),
              "updated": timezone.now(),
              "show_avatar": report.show_character_image
              }

    for f in fields:
        output['headers'][f.id] = {"rank": f.rank, "header": f.header,
                                   "field": f"field-{f.id}",
                                   "checkbox": f.checkbox_only,
                                   "aggregate": f.pass_fail_aggregate,
                                   "pass": 0,
                                   "description": f.long_descriptions,
                                   "allow_sort": f.allow_sort}

        try:  # TODO create new method for clean table data
            _f = f.report_Data_source.filter_object
            _pass = 0
            f_data = {}
            if hasattr(_f, "process_field"):
                f_data = _f.process_field(mains)
            elif hasattr(_f, "audit_filter"):
                logger.warning(
                    f"{_f} has no process_field, falling back to audit_filter")
                f_data = _f.audit_filter(mains)
            for u in mains:
                if u.id not in output['data']:
                    output['data'][u.id] = {"character": {
                        "name": u.profile.main_character.character_name, "id": u.profile.main_character.character_id}}
                output['data'][u.id][f"field-{f.id}"] = {
                    "check": f_data[u.id]['check'],
                    "data": "" if f.checkbox_only else f_data[u.id]['message']
                }
                output['headers'][f.id]['pass'] += 1 if f_data[u.id]['check'] else 0

        except Exception as e:
            logger.error(f"Error in Filter {e}")

    # strip the uid's from the data its just annoying to deal with.
    output['data'] = list(output['data'].values())
    output['headers'] = list(output['headers'].values())
    ReportResults.objects.update_or_create(corporation=corp, report=report,
                                           defaults={"results": json.dumps(output, cls=DjangoJSONEncoder)})

    return output



from __future__ import division

from datetime import datetime
from typing import Dict, List, Optional
from xmlrpc.client import Boolean, boolean

from ninja import Schema


class Message(Schema):
    message: str


class CharacterUpdate(Schema):
    data: str
    updated: datetime


class Character(Schema):
    character_name: str
    character_id: int
    corporation_id: int
    corporation_name: str
    alliance_id: Optional[int]
    alliance_name: Optional[str]


class Corporation(Schema):
    corporation_id: int
    corporation_name: str
    alliance_id: Optional[int]
    alliance_name: Optional[str]


class Report(Schema):
    name: str
    id: int


class MenuLink(Schema):
    name: str
    link: str


class MenuCategory(Schema):
    name: str
    links: List[MenuLink]


class CorpHistory(Schema):
    start: datetime
    corporation: Corporation


class CharacterHistory(Schema):
    character: Character
    history: List[CorpHistory] = None


class EveName(Schema):
    id: int
    name: str
    cat: Optional[str]



# Generated by Django 4.0.10 on 2023-06-16 13:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authstats', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportdatathrough',
            name='checkbox_only',
            field=models.BooleanField(default=False),
        ),
    ]



from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook
from django.utils.translation import gettext_lazy as _

from . import app_settings, models, urls


class AuthStats(MenuItemHook):
    def __init__(self):

        MenuItemHook.__init__(self,
                              app_settings.AUTHSTATS_APP_NAME,
                              'fas fa-book-open fa-fw',
                              'authstats:base',
                              navactive=['authstats:base'])

    def render(self, request):
        if request.user.has_perm('authstats.basic_access'):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return AuthStats()


@hooks.register('url_hook')
def register_url():
    return UrlHook(urls, 'authstats', r'^reports/')


@hooks.register("secure_group_filters")
def filters():
    return [models.ReportFieldAlts]


@hooks.register('discord_cogs_hook')
def register_cogs():
    return []



import json
import xml.etree.ElementTree as ET

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from django.contrib import messages
from django.contrib.auth.decorators import (login_required,
                                            permission_required,
                                            user_passes_test)
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django_celery_beat.models import CrontabSchedule, PeriodicTask
from esi.decorators import _check_callback, token_required
from esi.views import sso_redirect

from authstats.models import Report

from . import __version__

REQUIRED_SCOPES = [
    "esi-corporations.read_corporation_membership.v1"
]


@login_required
@token_required(scopes=REQUIRED_SCOPES)
def add_corp(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    corp, created = EveCorporationInfo.objects.get_or_create(corporation_id=char.corporation_id,
                                                             defaults={'member_count': 0,
                                                                       'corporation_ticker': char.corporation_ticker,
                                                                       'corporation_name': char.corporation_name
                                                                       })
    return redirect('authstats:base')


@permission_required("authstats.basic_access")
def react_main(request, rid, cid):
    # get available models
    return render(request, 'authstats/react_base.html', context={"version": __version__, "app_name": "authstats", "page_title": "Auth Reports"})


@permission_required("authstats.basic_access")
def react_redirect(request):
    # get available models
    rid = Report.objects.all().first().id
    cid = request.user.profile.main_character.corporation_id
    return redirect("authstats:report", rid=rid, cid=cid)



from django.contrib import admin
from django.utils.html import format_html
from solo.admin import SingletonModelAdmin

from .models import (AuthReportsConfiguration, Report, ReportDataSource,
                     ReportDataThrough, ReportFieldAlts)


class ReportDataFieldThroughAdmin(admin.StackedInline):
    model = ReportDataThrough


class ReportAdmin(admin.ModelAdmin):
    inlines = [
        ReportDataFieldThroughAdmin,
    ]


admin.site.register(Report, ReportAdmin)

admin.site.register(ReportDataSource)
admin.site.register(ReportFieldAlts)

admin.site.register(AuthReportsConfiguration, SingletonModelAdmin)



from django.apps import AppConfig

from . import __version__


class AuthStatsConfig(AppConfig):
    name = 'authstats'
    label = 'authstats'

    verbose_name = f"Auth Reports v{__version__}"

    def ready(self):
        import authstats.signals



import json

from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand

from ...models import Report
from ...signals import filters
from ...tasks import run_report_for_corp


class Command(BaseCommand):
    help = 'run a report'

    def add_arguments(self, parser):
        parser.add_argument('--report', type=int)
        parser.add_argument('--corp_id', type=int)

    def handle(self, *args, **options):
        self.stdout.write(
            f"Running a report for RID:{options['report']} CID:{options['corp_id']}")
        return json.dumps(run_report_for_corp(options['corp_id'], options['report']), indent=4)



# Generated by Django 4.0.10 on 2023-06-24 10:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('authstats', '0007_reportdatathrough_allow_sort'),
    ]

    operations = [
        migrations.AlterField(
            model_name='reportdatathrough',
            name='long_descriptions',
            field=models.CharField(
                blank=True, default=None, help_text='Long description shown to the user on hover of column name.', max_length=500, null=True),
        ),
    ]



import json

from django.core.management.base import BaseCommand

from ...models import ReportResults
from ...tasks import run_report_for_corp


class Command(BaseCommand):
    help = 'refresh all a reports'

    def handle(self, *args, **options):
        for rep in ReportResults.objects.all():
            self.stdout.write(
                f"Refreshing Report:    Corp:{rep.corporation}     Report:{rep.report.name}")
            run_report_for_corp.delay(
                rep.corporation.corporation_id, rep.report_id)
